@@ -18,9 +18,10 @@ from ipaddress import (
 from json import dumps, loads
 from uuid import UUID
 
-from dateutil.parser import parse
+from dateutil.parser import ParserError, parse
 
 from pg8000.exceptions import InterfaceError
+from pg8000.types import PGInterval, Range
 
 
 ANY_ARRAY = 2277
@@ -38,11 +39,23 @@ CSTRING = 2275
 CSTRING_ARRAY = 1263
 DATE = 1082
 DATE_ARRAY = 1182
+DATEMULTIRANGE = 4535
+DATEMULTIRANGE_ARRAY = 6155
+DATERANGE = 3912
+DATERANGE_ARRAY = 3913
 FLOAT = 701
 FLOAT_ARRAY = 1022
 INET = 869
 INET_ARRAY = 1041
 INT2VECTOR = 22
+INT4MULTIRANGE = 4451
+INT4MULTIRANGE_ARRAY = 6150
+INT4RANGE = 3904
+INT4RANGE_ARRAY = 3905
+INT8MULTIRANGE = 4536
+INT8MULTIRANGE_ARRAY = 6157
+INT8RANGE = 3926
+INT8RANGE_ARRAY = 3927
 INTEGER = 23
 INTEGER_ARRAY = 1007
 INTERVAL = 1186
@@ -59,11 +72,16 @@ NAME = 19
 NAME_ARRAY = 1003
 NUMERIC = 1700
 NUMERIC_ARRAY = 1231
+NUMRANGE = 3906
+NUMRANGE_ARRAY = 3907
+NUMMULTIRANGE = 4532
+NUMMULTIRANGE_ARRAY = 6151
 NULLTYPE = -1
 OID = 26
 POINT = 600
 REAL = 700
 REAL_ARRAY = 1021
+RECORD = 2249
 SMALLINT = 21
 SMALLINT_ARRAY = 1005
 SMALLINT_VECTOR = 22
@@ -76,6 +94,14 @@ TIMESTAMP = 1114
 TIMESTAMP_ARRAY = 1115
 TIMESTAMPTZ = 1184
 TIMESTAMPTZ_ARRAY = 1185
+TSMULTIRANGE = 4533
+TSMULTIRANGE_ARRAY = 6152
+TSRANGE = 3908
+TSRANGE_ARRAY = 3909
+TSTZMULTIRANGE = 4534
+TSTZMULTIRANGE_ARRAY = 6153
+TSTZRANGE = 3910
+TSTZRANGE_ARRAY = 3911
 UNKNOWN = 705
 UUID_TYPE = 2950
 UUID_ARRAY = 2951
@@ -117,7 +143,11 @@ def date_in(data):
     if data in ("infinity", "-infinity"):
         return data
     else:
-        return Datetime.strptime(data, "%Y-%m-%d").date()
+        try:
+            return Datetime.strptime(data, "%Y-%m-%d").date()
+        except ValueError:
+            # pg date can overflow Python Datetime
+            return data
 
 
 def date_out(v):
@@ -187,6 +217,10 @@ def numeric_out(d):
     return str(d)
 
 
+def point_in(data):
+    return tuple(map(float, data[1:-1].split(",")))
+
+
 def pg_interval_in(data):
     return PGInterval.from_str(data)
 
@@ -195,6 +229,17 @@ def pg_interval_out(v):
     return str(v)
 
 
+def range_out(v):
+    if v.is_empty:
+        return "empty"
+    else:
+        le = v.lower
+        val_lower = "" if le is None else make_param(PY_TYPES, le)
+        ue = v.upper
+        val_upper = "" if ue is None else make_param(PY_TYPES, ue)
+        return f"{v.bounds[0]}{val_lower},{val_upper}{v.bounds[1]}"
+
+
 def string_in(data):
     return data
 
@@ -220,7 +265,11 @@ def timestamp_in(data):
         pattern = "%Y-%m-%d %H:%M:%S.%f" if "." in data else "%Y-%m-%d %H:%M:%S"
         return Datetime.strptime(data, pattern)
     except ValueError:
-        return parse(data)
+        try:
+            return parse(data)
+        except ParserError:
+            # pg timestamp can overflow Python Datetime
+            return data
 
 
 def timestamptz_in(data):
@@ -231,7 +280,11 @@ def timestamptz_in(data):
         patt = "%Y-%m-%d %H:%M:%S.%f%z" if "." in data else "%Y-%m-%d %H:%M:%S%z"
         return Datetime.strptime(f"{data}00", patt)
     except ValueError:
-        return parse(data)
+        try:
+            return parse(data)
+        except ParserError:
+            # pg timestamptz can overflow Python Datetime
+            return data
 
 
 def unknown_out(v):
@@ -250,275 +303,66 @@ def uuid_in(data):
     return UUID(data)
 
 
-class PGInterval:
-    UNIT_MAP = {
-        "millennia": "millennia",
-        "millenium": "millennia",
-        "centuries": "centuries",
-        "century": "centuries",
-        "decades": "decades",
-        "decade": "decades",
-        "years": "years",
-        "year": "years",
-        "months": "months",
-        "month": "months",
-        "mon": "months",
-        "mons": "months",
-        "weeks": "weeks",
-        "week": "weeks",
-        "days": "days",
-        "day": "days",
-        "hours": "hours",
-        "hour": "hours",
-        "minutes": "minutes",
-        "minute": "minutes",
-        "mins": "minutes",
-        "secs": "seconds",
-        "seconds": "seconds",
-        "second": "seconds",
-        "microseconds": "microseconds",
-        "microsecond": "microseconds",
-    }
+def _range_in(elem_func):
+    def range_in(data):
+        if data == "empty":
+            return Range(is_empty=True)
+        else:
+            le, ue = [None if v == "" else elem_func(v) for v in data[1:-1].split(",")]
+            return Range(le, ue, bounds=f"{data[0]}{data[-1]}")
 
-    ISO_LOOKUP = {
-        True: {
-            "Y": "years",
-            "M": "months",
-            "D": "days",
-        },
-        False: {
-            "H": "hours",
-            "M": "minutes",
-            "S": "seconds",
-        },
-    }
+    return range_in
 
-    @classmethod
-    def from_str_iso_8601(cls, interval_str):
-        # P[n]Y[n]M[n]DT[n]H[n]M[n]S
-        kwargs = {}
-        lookup = cls.ISO_LOOKUP[True]
+
+daterange_in = _range_in(date_in)
+int4range_in = _range_in(int)
+int8range_in = _range_in(int)
+numrange_in = _range_in(Decimal)
+
+
+def ts_in(data):
+    return timestamp_in(data[1:-1])
+
+
+def tstz_in(data):
+    return timestamptz_in(data[1:-1])
+
+
+tsrange_in = _range_in(ts_in)
+tstzrange_in = _range_in(tstz_in)
+
+
+def _multirange_in(adapter):
+    def f(data):
+        in_range = False
+        result = []
         val = []
-
-        for c in interval_str[1:]:
-            if c == "T":
-                lookup = cls.ISO_LOOKUP[False]
-            elif c.isdigit() or c in ("-", "."):
+        for c in data:
+            if in_range:
                 val.append(c)
-            else:
-                val_str = "".join(val)
-                name = lookup[c]
-                v = float(val_str) if name == "seconds" else int(val_str)
-                kwargs[name] = v
-                val.clear()
+                if c in "])":
+                    value = "".join(val)
+                    val.clear()
+                    result.append(adapter(value))
+                    in_range = False
+            elif c in "[(":
+                val.append(c)
+                in_range = True
 
-        return cls(**kwargs)
+        return result
 
-    @classmethod
-    def from_str_postgres(cls, interval_str):
-        """Parses both the postgres and postgres_verbose formats"""
-
-        t = {}
-
-        curr_val = None
-        for k in interval_str.split():
-            if ":" in k:
-                hours_str, minutes_str, seconds_str = k.split(":")
-                hours = int(hours_str)
-                if hours != 0:
-                    t["hours"] = hours
-                minutes = int(minutes_str)
-                if minutes != 0:
-                    t["minutes"] = minutes
-
-                seconds = float(seconds_str)
-
-                if seconds != 0:
-                    t["seconds"] = seconds
-
-            elif k == "@":
-                continue
-
-            elif k == "ago":
-                for k, v in tuple(t.items()):
-                    t[k] = -1 * v
-
-            else:
-                try:
-                    curr_val = int(k)
-                except ValueError:
-                    t[cls.UNIT_MAP[k]] = curr_val
-
-        return cls(**t)
-
-    @classmethod
-    def from_str_sql_standard(cls, interval_str):
-        """YYYY-MM
-        or
-        DD HH:MM:SS.F
-        or
-        YYYY-MM DD HH:MM:SS.F
-        """
-        month_part = None
-        day_parts = None
-        parts = interval_str.split()
-
-        if len(parts) == 1:
-            month_part = parts[0]
-        elif len(parts) == 2:
-            day_parts = parts
-        else:
-            month_part = parts[0]
-            day_parts = parts[1:]
-
-        kwargs = {}
-
-        if month_part is not None:
-            if month_part.startswith("-"):
-                sign = -1
-                p = month_part[1:]
-            else:
-                sign = 1
-                p = month_part
-
-            kwargs["years"], kwargs["months"] = [int(v) * sign for v in p.split("-")]
-
-        if day_parts is not None:
-            kwargs["days"] = int(day_parts[0])
-            time_part = day_parts[1]
-
-            if time_part.startswith("-"):
-                sign = -1
-                p = time_part[1:]
-            else:
-                sign = 1
-                p = time_part
-
-            hours, minutes, seconds = p.split(":")
-            kwargs["hours"] = int(hours) * sign
-            kwargs["minutes"] = int(minutes) * sign
-            kwargs["seconds"] = float(seconds) * sign
-
-        return cls(**kwargs)
-
-    @classmethod
-    def from_str(cls, interval_str):
-        if interval_str.startswith("P"):
-            return cls.from_str_iso_8601(interval_str)
-        elif interval_str.startswith("@"):
-            return cls.from_str_postgres(interval_str)
-        else:
-            parts = interval_str.split()
-            if (len(parts) > 1 and parts[1][0].isalpha()) or (
-                len(parts) == 1 and ":" in parts[0]
-            ):
-                return cls.from_str_postgres(interval_str)
-            else:
-                return cls.from_str_sql_standard(interval_str)
-
-    def __init__(
-        self,
-        millennia=None,
-        centuries=None,
-        decades=None,
-        years=None,
-        months=None,
-        weeks=None,
-        days=None,
-        hours=None,
-        minutes=None,
-        seconds=None,
-        microseconds=None,
-    ):
-        self.millennia = millennia
-        self.centuries = centuries
-        self.decades = decades
-        self.years = years
-        self.months = months
-        self.weeks = weeks
-        self.days = days
-        self.hours = hours
-        self.minutes = minutes
-        self.seconds = seconds
-        self.microseconds = microseconds
-
-    def __repr__(self):
-        return f"<PGInterval {self}>"
-
-    def _value_dict(self):
-        return {
-            k: v
-            for k, v in (
-                ("millennia", self.millennia),
-                ("centuries", self.centuries),
-                ("decades", self.decades),
-                ("years", self.years),
-                ("months", self.months),
-                ("weeks", self.weeks),
-                ("days", self.days),
-                ("hours", self.hours),
-                ("minutes", self.minutes),
-                ("seconds", self.seconds),
-                ("microseconds", self.microseconds),
-            )
-            if v is not None
-        }
-
-    def __str__(self):
-        return " ".join(f"{v} {n}" for n, v in self._value_dict().items())
-
-    def normalize(self):
-        months = 0
-        if self.months is not None:
-            months += self.months
-        if self.years is not None:
-            months += self.years * 12
-
-        days = 0
-        if self.days is not None:
-            days += self.days
-        if self.weeks is not None:
-            days += self.weeks * 7
-
-        seconds = 0
-        if self.hours is not None:
-            seconds += self.hours * 60 * 60
-        if self.minutes is not None:
-            seconds += self.minutes * 60
-        if self.seconds is not None:
-            seconds += self.seconds
-        if self.microseconds is not None:
-            seconds += self.microseconds / 1000000
-
-        return PGInterval(months=months, days=days, seconds=seconds)
-
-    def __eq__(self, other):
-        if isinstance(other, PGInterval):
-            s = self.normalize()
-            o = other.normalize()
-            return s.months == o.months and s.days == o.days and s.seconds == o.seconds
-        else:
-            return False
-
-    def to_timedelta(self):
-        pairs = self._value_dict()
-        overlap = pairs.keys() & {
-            "weeks",
-            "months",
-            "years",
-            "decades",
-            "centuries",
-            "millennia",
-        }
-        if len(overlap) > 0:
-            raise ValueError(
-                "Can't fit the interval fields {overlap} into a datetime.timedelta."
-            )
-
-        return Timedelta(**pairs)
+    return f
 
 
-class ArrayState(Enum):
+datemultirange_in = _multirange_in(daterange_in)
+int4multirange_in = _multirange_in(int4range_in)
+int8multirange_in = _multirange_in(int8range_in)
+nummultirange_in = _multirange_in(numrange_in)
+tsmultirange_in = _multirange_in(tsrange_in)
+tstzmultirange_in = _multirange_in(tstzrange_in)
+
+
+class ParserState(Enum):
     InString = 1
     InEscape = 2
     InValue = 3
@@ -526,19 +370,19 @@ class ArrayState(Enum):
 
 
 def _parse_array(data, adapter):
-    state = ArrayState.Out
+    state = ParserState.Out
     stack = [[]]
     val = []
     for c in data:
-        if state == ArrayState.InValue:
+        if state == ParserState.InValue:
             if c in ("}", ","):
                 value = "".join(val)
                 stack[-1].append(None if value == "NULL" else adapter(value))
-                state = ArrayState.Out
+                state = ParserState.Out
             else:
                 val.append(c)
 
-        if state == ArrayState.Out:
+        if state == ParserState.Out:
             if c == "{":
                 a = []
                 stack[-1].append(a)
@@ -549,22 +393,22 @@ def _parse_array(data, adapter):
                 pass
             elif c == '"':
                 val = []
-                state = ArrayState.InString
+                state = ParserState.InString
             else:
                 val = [c]
-                state = ArrayState.InValue
+                state = ParserState.InValue
 
-        elif state == ArrayState.InString:
+        elif state == ParserState.InString:
             if c == '"':
                 stack[-1].append(adapter("".join(val)))
-                state = ArrayState.Out
+                state = ParserState.Out
             elif c == "\\":
-                state = ArrayState.InEscape
+                state = ParserState.InEscape
             else:
                 val.append(c)
-        elif state == ArrayState.InEscape:
+        elif state == ParserState.InEscape:
             val.append(c)
-            state = ArrayState.InString
+            state = ParserState.InString
 
     return stack[0][0]
 
@@ -580,16 +424,28 @@ bool_array_in = _array_in(bool_in)
 bytes_array_in = _array_in(bytes_in)
 cidr_array_in = _array_in(cidr_in)
 date_array_in = _array_in(date_in)
+datemultirange_array_in = _array_in(datemultirange_in)
+daterange_array_in = _array_in(daterange_in)
 inet_array_in = _array_in(inet_in)
 int_array_in = _array_in(int)
+int4multirange_array_in = _array_in(int4multirange_in)
+int4range_array_in = _array_in(int4range_in)
+int8multirange_array_in = _array_in(int8multirange_in)
+int8range_array_in = _array_in(int8range_in)
 interval_array_in = _array_in(interval_in)
 json_array_in = _array_in(json_in)
 float_array_in = _array_in(float)
 numeric_array_in = _array_in(numeric_in)
+nummultirange_array_in = _array_in(nummultirange_in)
+numrange_array_in = _array_in(numrange_in)
 string_array_in = _array_in(string_in)
 time_array_in = _array_in(time_in)
 timestamp_array_in = _array_in(timestamp_in)
 timestamptz_array_in = _array_in(timestamptz_in)
+tsrange_array_in = _array_in(tsrange_in)
+tsmultirange_array_in = _array_in(tsmultirange_in)
+tstzmultirange_array_in = _array_in(tstzmultirange_in)
+tstzrange_array_in = _array_in(tstzrange_in)
 uuid_array_in = _array_in(uuid_in)
 
 
@@ -615,10 +471,12 @@ def array_string_escape(v):
 def array_out(ar):
     result = []
     for v in ar:
-
-        if isinstance(v, (list, tuple)):
+        if isinstance(v, list):
             val = array_out(v)
 
+        elif isinstance(v, tuple):
+            val = f'"{composite_out(v)}"'
+
         elif v is None:
             val = "NULL"
 
@@ -636,7 +494,76 @@ def array_out(ar):
 
         result.append(val)
 
-    return "{" + ",".join(result) + "}"
+    return f'{{{",".join(result)}}}'
+
+
+def composite_out(ar):
+    result = []
+    for v in ar:
+        if isinstance(v, list):
+            val = array_out(v)
+
+        elif isinstance(v, tuple):
+            val = composite_out(v)
+
+        elif v is None:
+            val = ""
+
+        elif isinstance(v, dict):
+            val = array_string_escape(json_out(v))
+
+        elif isinstance(v, (bytes, bytearray)):
+            val = f'"\\{bytes_out(v)}"'
+
+        elif isinstance(v, str):
+            val = array_string_escape(v)
+
+        else:
+            val = make_param(PY_TYPES, v)
+
+        result.append(val)
+
+    return f'({",".join(result)})'
+
+
+def record_in(data):
+    state = ParserState.Out
+    results = []
+    val = []
+    for c in data:
+        if state == ParserState.InValue:
+            if c in (")", ","):
+                value = "".join(val)
+                val.clear()
+                results.append(None if value == "" else value)
+                state = ParserState.Out
+            else:
+                val.append(c)
+
+        if state == ParserState.Out:
+            if c in "(),":
+                pass
+            elif c == '"':
+                state = ParserState.InString
+            else:
+                val.append(c)
+                state = ParserState.InValue
+
+        elif state == ParserState.InString:
+            if c == '"':
+                results.append("".join(val))
+                val.clear()
+                state = ParserState.Out
+            elif c == "\\":
+                state = ParserState.InEscape
+            else:
+                val.append(c)
+
+        elif state == ParserState.InEscape:
+            val.append(c)
+            state = ParserState.InString
+
+    return tuple(results)
 
 
 PY_PG = {
@@ -670,6 +597,7 @@ PY_TYPES = {
     IPv4Network: inet_out,  # inet
     IPv6Network: inet_out,  # inet
     PGInterval: interval_out,  # interval
+    Range: range_out,  # range types
     Time: time_out,  # time
     Timedelta: interval_out,  # interval
     UUID: uuid_out,  # uuid
@@ -682,7 +610,7 @@ PY_TYPES = {
     str: string_out,  # unknown
     int: int_out,
     list: array_out,
-    tuple: array_out,
+    tuple: composite_out,
 }
 
 
@@ -700,10 +628,22 @@ PG_TYPES = {
     CSTRING_ARRAY: string_array_in,  # cstring[]
     DATE: date_in,  # date
     DATE_ARRAY: date_array_in,  # date[]
+    DATEMULTIRANGE: datemultirange_in,  # datemultirange
+    DATEMULTIRANGE_ARRAY: datemultirange_array_in,  # datemultirange[]
+    DATERANGE: daterange_in,  # daterange
+    DATERANGE_ARRAY: daterange_array_in,  # daterange[]
     FLOAT: float,  # float8
     FLOAT_ARRAY: float_array_in,  # float8[]
     INET: inet_in,  # inet
     INET_ARRAY: inet_array_in,  # inet[]
+    INT4MULTIRANGE: int4multirange_in,  # int4multirange
+    INT4MULTIRANGE_ARRAY: int4multirange_array_in,  # int4multirange[]
+    INT4RANGE: int4range_in,  # int4range
+    INT4RANGE_ARRAY: int4range_array_in,  # int4range[]
+    INT8MULTIRANGE: int8multirange_in,  # int8multirange
+    INT8MULTIRANGE_ARRAY: int8multirange_array_in,  # int8multirange[]
+    INT8RANGE: int8range_in,  # int8range
+    INT8RANGE_ARRAY: int8range_array_in,  # int8range[]
     INTEGER: int,  # int4
     INTEGER_ARRAY: int_array_in,  # int4[]
     JSON: json_in,  # json
@@ -717,11 +657,17 @@ PG_TYPES = {
     NAME_ARRAY: string_array_in,  # name[]
     NUMERIC: numeric_in,  # numeric
     NUMERIC_ARRAY: numeric_array_in,  # numeric[]
+    NUMRANGE: numrange_in,  # numrange
+    NUMRANGE_ARRAY: numrange_array_in,  # numrange[]
+    NUMMULTIRANGE: nummultirange_in,  # nummultirange
+    NUMMULTIRANGE_ARRAY: nummultirange_array_in,  # nummultirange[]
     OID: int,  # oid
+    POINT: point_in,  # point
     INTERVAL: interval_in,  # interval
     INTERVAL_ARRAY: interval_array_in,  # interval[]
     REAL: float,  # float4
     REAL_ARRAY: float_array_in,  # float4[]
+    RECORD: record_in,  # record
     SMALLINT: int,  # int2
     SMALLINT_ARRAY: int_array_in,  # int2[]
     SMALLINT_VECTOR: vector_in,  # int2vector
@@ -734,6 +680,14 @@ PG_TYPES = {
     TIMESTAMP_ARRAY: timestamp_array_in,  # timestamp
     TIMESTAMPTZ: timestamptz_in,  # timestamptz
     TIMESTAMPTZ_ARRAY: timestamptz_array_in,  # timestamptz
+    TSMULTIRANGE: tsmultirange_in,  # tsmultirange
+    TSMULTIRANGE_ARRAY: tsmultirange_array_in,  # tsmultirange[]
+    TSRANGE: tsrange_in,  # tsrange
+    TSRANGE_ARRAY: tsrange_array_in,  # tsrange[]
+    TSTZMULTIRANGE: tstzmultirange_in,  # tstzmultirange
+    TSTZMULTIRANGE_ARRAY: tstzmultirange_array_in,  # tstzmultirange[]
+    TSTZRANGE: tstzrange_in,  # tstzrange
+    TSTZRANGE_ARRAY: tstzrange_array_in,  # tstzrange[]
     UNKNOWN: string_in,  # unknown
     UUID_ARRAY: uuid_array_in,  # uuid[]
     UUID_TYPE: uuid_in,  # uuid
@@ -852,14 +806,16 @@ def literal(value):
         return str(value)
     elif isinstance(value, (bytes, bytearray)):
         return f"X'{value.hex()}'"
+    elif isinstance(value, Datetime):
+        return f"'{datetime_out(value)}'"
     elif isinstance(value, Date):
         return f"'{date_out(value)}'"
     elif isinstance(value, Time):
         return f"'{time_out(value)}'"
-    elif isinstance(value, Datetime):
-        return f"'{datetime_out(value)}'"
     elif isinstance(value, Timedelta):
         return f"'{interval_out(value)}'"
+    elif isinstance(value, list):
+        return f"'{array_out(value)}'"
     else:
         val = str(value).replace("'", "''")
         return f"'{val}'"
